/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2001 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

#define APR_WANT_STRFUNC /* for strncmp, strlen, etc */
#define APR_WANT_MEMFUNC /* for memset, etc */
#define APR_WANT_STDIO   /* for sscanf */
#include <apr_want.h>

#include "apr_portable.h"
#include "apr_time.h"
#include "apr_lib.h"
#include "apr_private.h"

/* an implementation using strptime(3) for platforms which have it */
#if defined(HAVE_STRPTIME) && (APR_HAVE_SYS_TIME_H || defined(HAVE_TIME_H))

/* System Headers required for time library */
#if APR_HAVE_SYS_TIME_H
# include <sys/time.h>
#endif
#ifdef HAVE_TIME_H
# include <time.h>
#endif

/* this works iff:
 *  a) system has (working) strptime function
 *     (even some non-working strptime work, see implementation)
 *  b) apr_os_exp_time is same as struct tm
 *  c) apr_os_exp_time_put ignores pool argument. */
/* XXX: how to enforce the above? */
apr_status_t apr_strptime(const char *s, char **retptr,
			  const char *format, apr_exploded_time_t *tm)
{
  struct tm os_tm, *os_tmp = &os_tm;
  apr_status_t status;
  status = apr_os_exp_time_get(&os_tmp, tm);
  if (status != APR_SUCCESS) goto bail;
  (*retptr) = strptime(s, format, os_tmp);
  if (NULL == (*retptr)) {
    status = APR_EBADDATE;
    goto bail;
  }
  /* versions of glibc before 10-aug-2001 contain bug parsing %OI */
  /* see cvs log for revision 1.34 of libc/time/strptime.c */
  /* so test for this bug and work-around it if necessary */
  {
    struct tm test = os_tm;
    test.tm_hour=0; /* protect against *really* broken strptime */
    strptime ("8", "%OI", &test);
    if (test.tm_hour == 7) /* BUGGY GLIBC! */
      {                         
	const char *cp;
	/* first determine whether the format string contains %OI,
	 * necessitating a fixup. */
	int needs_fixup=0;
	for (cp = format; *cp && !needs_fixup; cp++)
	  if (*cp=='%' && *++cp=='O' && *++cp=='I')
	    needs_fixup=1;
	if (needs_fixup)
	  {
	    /* fixup the bug, sigh. */
	    os_tm.tm_hour = os_tm.tm_hour + 1;
	    if (os_tm.tm_hour == 12 || os_tm.tm_hour == 24)
	      os_tm.tm_hour -= 12;
	  }
      }
  }
  /* -- end bug workaround -- */
  status = apr_os_exp_time_put(tm, &os_tmp, NULL/*craaazy*/);
 bail:
  return status;
}

#else /* use the non-strptime implementation.  limited features, no l18n */

/* table of 'first day of month' for normal and leap years */
const int mon_yday[2][13] = {
    /* Normal years.  */
    { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365 },
    /* Leap years.  */
    { 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366 }
};

/* date routines */
static int is_leap_year(int year) {
  return year % 4 == 0 && (year % 100 != 0 || year % 400 == 0);
}
static void month_and_day(apr_exploded_time_t *tm) {
  int mon = 0, isleap = is_leap_year(1900+tm->tm_year);
  while (mon<12 && mon_yday[isleap][mon+1] < tm->tm_yday)
    mon++;
  tm->tm_mon = mon;
  tm->tm_mday = (tm->tm_yday - mon_yday[isleap][mon]) + 1;
}
static void day_of_the_week(apr_exploded_time_t *tm) {
  /* deep magic, borrowed from glibc strptime */
  /* We know that January 1st 1970 was a Thursday (= 4).  Compute the
     the difference between this data in the one on TM and so determine
     the weekday.  */
  int corr_year = 1900 + tm->tm_year - (tm->tm_mon < 2);
  int wday = (-473
	      + (365 * (tm->tm_year - 70))
	      + (corr_year / 4)
	      - ((corr_year / 4) / 25) + ((corr_year / 4) % 25 < 0)
	      + (((corr_year / 4) / 25) / 4)
	      + mon_yday[0][tm->tm_mon]
	      + tm->tm_mday - 1);
  tm->tm_wday = ((wday % 7) + 7) % 7;
}
static void day_of_the_year(apr_exploded_time_t *tm) {
  tm->tm_yday = (mon_yday[is_leap_year (1900 + tm->tm_year)][tm->tm_mon]
		 + (tm->tm_mday - 1));
}

/* Lists of day/month/am/pm names, with associated integer values */
typedef struct {
  char *phrase; int val;
} string_list_t;
const static string_list_t weekday_list[] = {
  { "Sunday",    0 },  { "Sun",  0 },
  { "Monday",    1 },  { "Mon",  1 },
  { "Tuesday",   2 },  { "Tue",  2 },
  { "Wednesday", 3 },  { "Wed",  3 },
  { "Thursday",  4 },  { "Thu",  4 },
  { "Friday",    5 },  { "Fri",  5 },
  { "Saturday",  6 },  { "Sat",  6 },
  { NULL, 0 }
};
const static string_list_t month_list[] = {
  { "January",   0 },  { "Jan",  0 },
  { "February",  1 },  { "Feb",  1 },
  { "March",     2 },  { "Mar",  2 },
  { "April",     3 },  { "Apr",  3 },
  { "May",       4 },  { "May",  4 },
  { "June",      5 },  { "Jun",  5 },
  { "July",      6 },  { "Jul",  6 },
  { "August",    7 },  { "Aug",  7 },
  { "September", 8 },  { "Sep",  8 },
  { "October",   9 },  { "Oct",  9 },
  { "November", 10 },  { "Nov", 10 },
  { "December", 11 },  { "Dec", 11 },
  { NULL, 0 }
};
const static string_list_t am_pm_list[] = {
  { "AM", 0 },  { "A.M.", 0 },
  { "PM", 1 },  { "P.M.", 1 },
  { NULL, 0 }
};
    
/* Parse a string from the given string_list_t array.  Puts the
 * associated value in 'retval' and returns the first character
 * not in the string if found in 'sl'; or else returns NULL if
 * the string is not found in the list. */
static const char *get_string(const char *s, int *retval,
			      const string_list_t *sl) {
  int i;
  for (i=0; sl[i].phrase!=NULL; i++)
    if (strncasecmp(s, sl[i].phrase, strlen(sl[i].phrase))==0) {
      *retval = sl[i].val;
      return s+strlen(sl[i].phrase);
    }
  return NULL;
}

/* scan a number between 'min' and 'max' which is at most 'width'
   digits long. returns a pointer to the first character after the
   number if successful, or NULL if not. */
static const char *get_number(const char *s, int *retval,
			      int min, int max, int width) {
  const char *cp = s;
  int val = 0;
  /* must start with a number */
  if (!(*cp >= '0' && *cp <= '9'))
    return NULL;
  /* first number */
  val = (*cp++)-'0';
  if (val>max) return NULL; /* can't even get one digit w/o being too large */
  /* get digits until: none left, max width, or value is too large */
  while (*cp >= '0' && *cp <= '9' && (cp-s) < width && val < max)
    val = (val*10)+((*cp++)-'0');
  /* we've gotten all the digits; let's make sure it's valid */
  if (val>max) {
    /* if value is too large, try to fix it up by 'ungetting' the last digit */
    val/=10; cp--;
  }
  /* if value is still out of range, return NULL */
  if (val<min || val>max) return NULL;
  /* okay, we've got a good number. */
  *retval = val;
  return cp;
}

/* internal state of the parser, preserved even if we invoke
 * apr_strptime_internal recursively. We need this information in
 * order to fix up certain fields at the end of the parse. */
typedef struct {
  int got_12hour, got_ampm, got_wday, got_mday, got_mon, got_yday, got_year;
  int got_year_in_century, got_century;
  int is_pm, year_in_century, century;
} internal_state_t;

static const char *apr_strptime_internal(const char *s, const char *format,
					 apr_exploded_time_t *tm,
					 internal_state_t *state) {
  int i;
  while (*format!='\0') {
    /* spaces in the format string match 0 or more spaces in the input */
    if (apr_isspace(*format)) {
      while (apr_isspace(*s))
	s++;
      format++;
      continue;
    }
    /* any character other than '%' must match itself. */
    if (*format!='%') {
      if (*s!=*format)
	return NULL; /* no match */
      format++; s++;
      continue;
    }
    /* okay, we've got a '%' field descriptor.  Advance past the '%' and
     * switch on the field descriptor. */
    format++;
    if (*format=='E' || *format=='O')
      format++; /* skip 'locale-dependent' field modifier */
    switch(*format++) {
    case '%':
      if (*s++!='%') return NULL;
      break;
      /* recursive fields: these descriptors are just abbreviations of
       * others.  We invoke the parser recursively to handle them. */
    case 'c':
      if (NULL == (s = apr_strptime_internal(s, "%x %X", tm, state)))
	return NULL;
      break;
    case 'D':
      if (NULL == (s = apr_strptime_internal(s, "%m/%d/%y", tm, state)))
	return NULL;
      break;
    case 'r':
      if (NULL == (s = apr_strptime_internal(s, "%I:%M:%S %p", tm, state)))
	return NULL;
      break;
    case 'R':
      if (NULL == (s = apr_strptime_internal(s, "%H:%M", tm, state)))
	return NULL;
      break;
    case 'T':
      if (NULL == (s = apr_strptime_internal(s, "%H:%M:%S", tm, state)))
	return NULL;
      break;
    case 'x':
      /* "locale's date format" is %D */
      if (NULL == (s = apr_strptime_internal(s, "%D", tm, state)))
	return NULL;
      break;
    case 'X':
      /* "locale's time format" is %T */
      if (NULL == (s = apr_strptime_internal(s, "%T", tm, state)))
	return NULL;
      break;
      /* -------- */
      /* Numeric field types.  We use the 'get_number' routine to
       * parse these from the input.  get_number grabs the largest
       * string from the input that satisfies its value constraints. */
    case 'C': /* century */
      if (NULL == (s = get_number(s, &i, 0, 99, 2)))
	return NULL;
      state->got_century = 1;
      state->century = i;
      break;
    case 'd':
    case 'e':
      if (NULL == (s = get_number(s, &i, 1, 31, 2)))
	return NULL;
      tm->tm_mday = i;
      state->got_mday = 1;
      break;
    case 'H':
    case 'k':
      if (NULL == (s = get_number(s, &i, 0, 23, 2)))
	return NULL;
      tm->tm_hour = i;
      state->got_12hour = 0;
      break;
    case 'I':
    case 'l':
      if (NULL == (s = get_number(s, &i, 0, 12, 2)))
	return NULL;
      tm->tm_hour = i % 12;
      state->got_12hour = 1;
      break;
    case 'j':
      if (NULL == (s = get_number(s, &i, 1, 366, 3)))
	return NULL;
      tm->tm_yday = i - 1;
      state->got_yday = 1;
      break;
    case 'm':
      if (NULL == (s = get_number(s, &i, 1, 12, 2)))
	return NULL;
      tm->tm_mon = i - 1;
      state->got_mon = 1;
      break;
    case 'M':
      if (NULL == (s = get_number(s, &i, 0, 59, 2)))
	return NULL;
      tm->tm_min = i;
      break;
    case 'S':
      /* despite what the Single Unix specification says, no minute
       * has more than one leap second.  Thus the range is 0-60
       * (not 0-61, as the spec says).  glibc backs me up on this. */
      if (NULL == (s = get_number(s, &i, 0, 60, 2)))
	return NULL;
      tm->tm_sec = i;
      break;
      /* No implementation for %U and %W, which are week numbers using
       * Sunday as the first day of the week and Monday as the first
       * day of the week, respectively.  They're not terribly useful,
       * don't correspond to fields in apr_exploded_time_t, and aren't
       * even useful for parsing the ISO 8601 format for dates, because
       * they're defined wrong.  So we don't support them. */
    case 'w':
      if (NULL == (s = get_number(s, &i, 0, 6, 1)))
	return NULL;
      tm->tm_wday = i;
      state->got_wday = 1;
      break;
    case 'y': /* 'year within century'.  when century not otherwise
	       * specified, refers to 1969-2068 */
      if (NULL == (s = get_number(s, &i, 0, 99, 2)))
	return NULL;
      state->year_in_century = i;
      state->got_year_in_century = 1;
      break;
    case 'Y': /* year including century */
      /* not y10k aware */
      if (NULL == (s = get_number(s, &i, 0, 9999, 4)))
	return NULL;
      tm->tm_year = i-1900;
      state->got_year_in_century = state->got_century = 0;
      state->got_year = 1;
      break;
      /* -------- */
      /* String field types.  We use 'get_string' with our lists
       * of possible input strings.  This would be much more
       * complicated if we were trying to make this routine
       * localizable, but we're not.  Use your system-defined
       * strptime(3) routine if you want localizability. */
    case 'A':
    case 'a':
      if (NULL == (s = get_string(s, &i, weekday_list)))
	return NULL;
      tm->tm_wday = i;
      state->got_wday = 1;
      break;
    case 'B':
    case 'b':
    case 'h':
      if (NULL == (s = get_string(s, &i, month_list)))
	return NULL;
      tm->tm_mon = i;
      state->got_mon = 1;
      break;
    case 'p':
      if (NULL == (s = get_string(s, &i, am_pm_list)))
	return NULL;
      state->is_pm = i;
      state->got_ampm = 1;
      break;
      /* -------- */
      /* Invalid field specified.  Return 'no match'. */
    default:
      return NULL;
    }
  }
  /* done parsing the format string! */
  return s;
}

apr_status_t apr_strptime(const char *s, char **retptr,
			  const char *format, apr_exploded_time_t *tm)
{
  const char *cp;
  internal_state_t state;
  memset(&state, 0, sizeof(state));

  cp = apr_strptime_internal(s, format, tm, &state);
  if (cp==NULL)
    return APR_EBADDATE;
  /* fixup year if year-in-century/century was given */
  if (state.got_year_in_century) {
    if (state.got_century)
      tm->tm_year = (state.century*100 + state.year_in_century) - 1900;
    else { /* without century designation, years 00-68 refer to 2000-2068 */
      tm->tm_year = state.year_in_century;
      if (tm->tm_year < 69) tm->tm_year+=100;
    }
    state.got_year = 1;
  }
  /* fix-up 12hour when am/pm is known */
  if (state.got_12hour && state.got_ampm) {
    if (state.is_pm) tm->tm_hour+=12;
  }
  /* fixup month and mday if yday and year are known */
  if (state.got_yday && state.got_year &&
      !(state.got_mon && state.got_mday)) {
    month_and_day(tm);
    state.got_mon = state.got_mday = 1;
  }
  /* fixup wday if enough info is present */
  if (state.got_mon && state.got_mday && state.got_year &&
      !state.got_wday) {
    day_of_the_week(tm);
    state.got_wday = 1;
  }
  /* fixup yday if enough info is present */
  if (state.got_mon && state.got_mday && state.got_year &&
      !state.got_yday) {
    day_of_the_year(tm);
    state.got_yday = 1;
  }
  /* done */
  *retptr = (char*) cp; /* cast necessary due to hole in type system */
  return APR_SUCCESS;
}

#endif /* !HAVE_STRPTIME */
